from django.apps import apps
from django.db import models
from django.template import Library

from ..helpers import URLHelper

register = Library()

_url_helper_registry = {}


@register.simple_tag(takes_context=True)
def object_url(context, namespace, instance, **kwargs):
    if isinstance(instance, models.Model):
        model = instance.__class__
    elif isinstance(instance, str):
        model = apps.get_model(instance)
    else:
        model = instance
    opts = model._meta
    slug = "%s_%s" % (opts.app_label, opts.model_name)
    helper = _url_helper_registry.get(slug, None)
    if not helper:
        helper = URLHelper(namespace, model)
        _url_helper_registry[slug] = helper
    return helper.get_url(**kwargs)


@register.simple_tag(takes_context=True)
def get_opts(context, instance):
    if not isinstance(instance, (models.Model,)):
        raise ValueError("'%s' is not model instance" % instance)
    return instance._meta


@register.filter
def dashboard_urlname(value, arg):
    if isinstance(value, (models.Model,)):
        value = value._meta
    return "dashboard_%s_%s_%s" % (value.app_label, value.model_name, arg)


@register.filter
def website_urlname(value, arg):
    if isinstance(value, (models.Model,)):
        value = value._meta
    return "website_%s_%s_%s" % (value.app_label, value.model_name, arg)



import json
from urllib.parse import unquote

from django.http import HttpRequest
from django.utils.functional import cached_property


class HtmxMiddleware:
    """Simply copied form Django HTMX"""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request: HttpRequest):
        request.htmx = HtmxDetails(request)
        return self.get_response(request)


class HtmxDetails:
    def __init__(self, request):
        self.request = request

    def _get_header_value(self, name):
        value = self.request.headers.get(name) or None
        if value:
            if self.request.headers.get(f"{name}-URI-AutoEncoded") == "true":
                value = unquote(value)
        return value

    def __bool__(self):
        return self._get_header_value("HX-Request") == "true"

    @cached_property
    def boosted(self):
        return self._get_header_value("HX-Boosted") == "true"

    @cached_property
    def current_url(self):
        return self._get_header_value("HX-Current-URL")

    @cached_property
    def history_restore_request(self):
        return self._get_header_value("HX-History-Restore-Request") == "true"

    @cached_property
    def prompt(self):
        return self._get_header_value("HX-Prompt")

    @cached_property
    def target(self):
        return self._get_header_value("HX-Target")

    @cached_property
    def trigger(self):
        return self._get_header_value("HX-Trigger")

    @cached_property
    def trigger_name(self):
        return self._get_header_value("HX-Trigger-Name")

    @cached_property
    def triggering_event(self):
        value = self._get_header_value("Triggering-Event")
        if value is not None:
            try:
                value = json.loads(value)
            except json.JSONDecodeError:
                value = None
        return value



from django.contrib.auth.decorators import login_required
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator
from django.db.models import Model
from django.urls import path, re_path
from django.utils.decorators import method_decorator
from django.utils.safestring import mark_safe

from django_tables2.tables import Table, table_factory

from .helpers import ButtonHelper, PermissionHelper, URLHelper
from .views import InspectView, ListView

login_required_m = method_decorator(login_required)


class BaseViewSet:
    icon = ""
    prefix = None
    menu_label = None
    menu_icon = None
    menu_order = None

    def __init__(self, router=None):
        """Don't allow initialisation unless self.model is set to a valid model"""
        self.router = router

    @property
    def urls(self):
        return self.get_urls()

    def get_prefix(self):
        return "%s" % self.prefix

    def get_menu_label(self):
        """Returns the label text to be used for the menu item."""
        return self.menu_label or self.opts.verbose_name_plural.title()

    def get_menu_icon(self):
        return self.menu_icon

    def get_menu_order(self):
        """
        Returns the 'order' to be applied to the menu item. 000 being first
        place. Where ViewSetGroup is used, the menu_order value should be
        applied to that, and any ModelAdmin classes added to 'items'
        attribute will be ordered automatically, based on their order in that
        sequence.
        """
        return self.menu_order or 999

    def get_urls(self):
        """Utilised by 'register_webapp_urls' hook to register urls to router."""
        return []

    def has_view(self, view_name):
        return hasattr(self, view_name)


class BaseFormViewset:

    form_view_extra_css = []
    form_view_extra_js = []
    form_fields_exclude = []
    prepopulated_fields = {}
    success_url_name = "index"

    def get_success_url(self):
        if (self.success_url_name == "index" and self.has_view("index_view")) or (
            self.success_url_name == "create" and self.has_view("index_view")
        ):
            return self.url_helper.get_action_url(self.success_url_name)
        if (self.success_url_name == "edit" and self.has_view("index_view")) or (
            self.success_url_name == "inspect" and self.has_view("index_view")
        ):
            return self.url_helper.get_action_url(self.success_url_name, pk=self.pk)
        else:
            raise ImproperlyConfigured("success_url_name `index`, `create`, `edit`, `inspect`")

    def get_form_view_extra_css(self):
        return self.form_view_extra_css

    def get_form_view_extra_js(self):
        return self.form_view_extra_js


class ModelViewSet(BaseViewSet):
    model = None
    # Helper class
    url_helper_class = URLHelper
    button_helper_class = ButtonHelper
    permission_helper_class = PermissionHelper
    template_namespace = None

    def __init__(self, router=None):
        """Don't allow initialisation unless self.model is set to a valid model"""
        super().__init__(router=router)
        if not self.model or not issubclass(self.model, Model):
            raise ImproperlyConfigured(
                "The model attribute on your '%s' class must be set, and "
                "must be a valid Django model." % self.__class__.__name__
            )
        self.opts = self.model._meta
        self.namespace = self.router.namespace
        self.url_helper = self.get_url_helper_class()(self.namespace, self.model)
        self.permission_helper = self.get_permission_helper_class()(self.model)

    def get_queryset(self):
        """Returns a QuerySet of all model instances."""
        qs = self.model._default_manager.get_queryset()
        return qs

    def get_permission_helper_class(self):
        """Returns a permission_helper class to help with permission-based logic."""
        return self.permission_helper_class

    def get_permission_helper(self):
        return self.permission_helper

    def get_button_helper_class(self):
        """Returns a ButtonHelper class to help generate buttons for the given model."""
        return self.button_helper_class

    def get_button_helper(self):
        return self.button_helper

    def get_url_helper_class(self):
        return self.url_helper_class

    def get_url_helper(self):
        return self.url_helper

    def get_prefix(self):
        return "%s/%s/" % (str(self.opts.app_label).replace("_", "-"), self.opts.model_name)

    def get_templates(self, action="index"):
        """
        Utility function that provides a list of templates to try for a given
        view, when the template isn't overridden by one of the template
        attributes on the class.
        """
        namespace = self.router.namespace
        app_label = self.opts.app_label.lower()
        model_name = self.opts.model_name.lower()
        templates = [
            "%s/%s/%s/%s.html" % (namespace, app_label, model_name, action),
            "%s/%s/%s_%s.html" % (namespace, app_label, model_name, action),
            "%s/%s/%s.html" % (namespace, model_name, action),
            "%s/%s_%s.html" % (namespace, model_name, action),
            "%s/%s.html" % (namespace, action),
        ]
        return templates


class ListViewSetMixin(ModelViewSet):
    index_public = True
    index_title = None
    index_view_extra_css = list()
    index_view_extra_js = list()
    index_view_class = ListView
    index_template_name = None

    paginate_by = 20
    paginate_orphans = 0
    paginator_class = Paginator
    filterset_fields = '__all__'
    filterset_class = None
    select_related = False
    ordering = ("pk",)

    def get_queryset(self):
        """
        Returns a QuerySet of all model instances that can be edited by the
        admin site.
        """
        qs = super().get_queryset()
        ordering = self.get_ordering()
        if ordering:
            qs = qs.order_by(*ordering)
        select_related = self.get_select_related()
        if select_related:
            qs.select_related(*select_related)
        return qs

    def get_filterset_class(self):
        """Return self.filterset class, if None load default FilterView"""
        return self.filterset_class

    def get_ordering(self):
        return self.ordering or ()

    def get_select_related(self):
        return self.select_related

    def get_paginate_by(self):
        return self.paginate_by

    def get_paginator_class(self):
        return self.paginator_class

    def get_index_view_extra_css(self):
        css = []
        css.extend(self.index_view_extra_css)
        return css

    def get_index_view_extra_js(self):
        return self.index_view_extra_js

    def get_filterset_fields(self):
        """
        Returns a sequence containing the fields to be displayed as filters in the right sidebar
        in the list view.
        """
        return self.filterset_fields

    def get_index_title(self):
        return self.index_title or self.opts.verbose_name_plural.title()

    def get_index_template(self):
        return self.index_template_name or self.get_templates("index")

    def index_view(self, request):
        self.request = request
        kwargs = {
            "viewset": self,
            "title": self.get_index_title(),
        }
        view_class = self.index_view_class
        return view_class.as_view(**kwargs)(request)

    def get_urls(self):
        """
        Append urls to generic viewsets.
        """
        urls = super().get_urls()
        urls = urls + [
            path(
                self.url_helper.get_pattern("index"),
                self.index_view,
                name=self.url_helper.get_name("index"),
            ),
        ]
        return urls


class InspectViewSetMixin(BaseViewSet):

    inspect_public = True
    inspect_view_fields = []
    inspect_view_fields_exclude = []
    inspect_view_extra_css = []
    inspect_view_extra_js = []
    inspect_view_class = InspectView
    inspect_template_name = None

    def inspect_view(self, request, pk):
        self.request = request
        kwargs = {
            "viewset": self,
            "title": self.get_inspect_title(),
        }
        view_class = self.inspect_view_class
        return view_class.as_view(**kwargs)(request, pk=pk)

    def get_inspect_title(self):
        return self.index_title or "%s Detail" % self.opts.verbose_name.title()

    def get_inspect_template(self):
        return self.inspect_template_name or self.get_templates("inspect")

    def get_inspect_view_extra_css(self):
        return self.inspect_view_extra_css

    def get_inspect_view_extra_js(self):
        return self.inspect_view_extra_js

    def get_inspect_view_fields(self):
        """
        Return a list of field names, indicating the model fields that
        should be displayed in the 'inspect' view. Returns the value of the
        'inspect_view_fields' attribute if populated, otherwise a sensible
        list of fields is generated automatically, with any field named in
        'inspect_view_fields_exclude' not being included.
        """
        if not self.inspect_view_fields:
            found_fields = []
            for f in self.model._meta.get_fields():
                if f.name not in self.inspect_view_fields_exclude:
                    if f.concrete and (not f.is_relation or (not f.auto_created and f.related_model)):
                        found_fields.append(f.name)
            return found_fields
        return self.inspect_view_fields

    def get_urls(self):
        """
        Append urls to generic viewsets.
        """
        urls = super().get_urls()
        urls = urls + [
            re_path(
                self.url_helper.get_pattern("inspect", specific=True),
                self.inspect_view,
                name=self.url_helper.get_name("inspect"),
            ),
        ]
        return urls


class TableViewSetMixin(ListViewSetMixin):

    list_export = tuple()
    table_class = None
    table_template = "shared/table.html"
    table_add_buttons = None
    list_display = None
    list_display_exclude = None
    empty_value_display = "-"

    def get_table_class(self, request):
        table_class = self.table_class
        if not table_class:
            table_class = table_factory(
                table=Table,
                model=self.model,
                fields=self.get_list_display(request),
                exclude=self.get_list_display_exclude(request),
            )
        return table_class

    def get_table_kwargs(self, request):
        """
        Return a table object to use. The table has automatic support for
        sorting and pagination.
        """
        return {"template_name": self.get_table_template()}

    def get_table_template(self):
        return self.table_template

    def get_list_display(self, request):
        """
        Return a sequence containing the fields/method output
        to be displayed in the column view.
        """
        return self.list_display

    def get_list_display_exclude(self, request):
        """
        Return a sequence containing the fields/method output
        to be displayed in the list view.
        """
        return self.list_display_exclude

    def get_list_display_add_buttons(self, request):
        """
        Return the name of the field/method from list_display where
        action buttons should be added. Defaults to the first
        item from get_list_display()
        """
        return self.list_display_add_buttons or self.get_list_display(request)[0]

    def get_list_export(self, request):
        """
        Return a sequence containing the fields/method output to be displayed in spreadsheet exports.
        """
        return self.list_export

    def get_empty_value_display(self, field_name=None):
        """Return the empty_value_display value defined on ModelAdmin"""
        return mark_safe(self.empty_value_display)

    def get_extra_attrs_for_row(self, obj, context):
        """
        Return a dictionary of HTML attributes to be added to the `<tr>`
        element for the suppled `obj` when rendering the results table in
        `index_view`. `data-object-pk` is already added by default.
        """
        return {}

    def get_extra_class_names_for_field_col(self, obj, field_name):
        """
        Return a list of additional CSS class names to be added to the table
        cell's `class` attribute when rendering the output of `field_name` for
        `obj` in `index_view`.

        Must always return a list.
        """
        return []

    def get_extra_attrs_for_field_col(self, obj, field_name):
        """
        Return a dictionary of additional HTML attributes to be added to a
        table cell when rendering the output of `field_name` for `obj` in
        `index_view`.

        Must always return a dictionary.
        """
        return {}


class ReadOnlyViewSet(InspectViewSetMixin, ListViewSetMixin):
    pass



from ..settings import routers_settings


class MaintenanceMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        from ..views import MaintenanceView

        if routers_settings.MAINTENANCE_MODE_FORCE:
            return MaintenanceView.as_view()(request).render()
        return self.get_response(request)



from django.urls import include
from django.urls.conf import path

from .routers import SimpleRouter


class AuthenticationRouter(SimpleRouter):
    namespace = "accounts"
    index_enabled = False

    def get_urls(self):
        urls = []
        try:
            from allauth.account import views as allviews

            urls += [
                path("signup/", allviews.signup, name="account_signup"),
                path("login/", allviews.login, name="account_login"),
                path("logout/", allviews.logout, name="account_logout"),
                path(
                    "%s/" % self.namespace,
                    include("simpel.simpel_auth.urls.account"),
                ),
            ]
        except ImportError as err:
            print(err)
        return urls



from django.contrib.admin.utils import quote as admin_quote
from django.urls import reverse
from django.utils.functional import cached_property


class URLHelper:
    def __init__(self, namespace, model):
        self.namespace = namespace
        self.model = model
        self.opts = model._meta

    def get_pattern(self, action, specific=False):
        if specific:
            return r"^%s/(?P<pk>[-\w]+)/$" % action
        else:
            return "" if action == "index" else r"^%s/$" % action

    def get_name(self, action):
        return "%s_%s_%s_%s" % (
            self.namespace,
            self.opts.app_label,
            self.opts.model_name,
            action,
        )

    def get_url(self, action, *args, **kwargs):
        if action in ("create", "index"):
            return reverse(self.get_name(action))
        url_name = self.get_name(action)
        return reverse(url_name, args=args, kwargs=kwargs)

    @cached_property
    def index_url(self):
        return self.get_url("index")

    @cached_property
    def create_url(self):
        return self.get_url("create")


# Subclasses should define url_helper and permission_helper
class URLFinder:
    def __init__(self, user):
        self.user = user

    def get_edit_url(self, instance):
        if self.permission_helper.user_can_edit_obj(self.user, instance):
            return self.url_helper.get_url("edit", admin_quote(instance.pk))



import json

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.http.response import HttpResponseRedirectBase

HTMX_STOP_POLLING = 286


class HttpResponseStopPolling(HttpResponse):
    status_code = HTMX_STOP_POLLING

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._reason_phrase = "Stop Polling"


class HttpResponseClientRedirect(HttpResponseRedirectBase):
    status_code = 200

    def __init__(self, redirect_to, *args, **kwargs):
        super().__init__(redirect_to, *args, **kwargs)
        self["HX-Redirect"] = self["Location"]
        del self["Location"]


def trigger_client_event(response, name, params, *, after="receive"):
    if after == "receive":
        header = "HX-Trigger"
    elif after == "settle":
        header = "HX-Trigger-After-Settle"
    elif after == "swap":
        header = "HX-Trigger-After-Swap"
    else:
        raise ValueError("Value for 'after' must be one of: 'receive', 'settle', or 'swap'.")

    if header in response:
        value = response[header]
        try:
            data = json.loads(value)
        except json.JSONDecodeError as exc:
            raise ValueError(f"{header!r} value should be valid JSON.") from exc
        data[name] = params
    else:
        data = {name: params}

    response[header] = json.dumps(data, cls=DjangoJSONEncoder)



from django.contrib.auth.decorators import login_required
from django.urls import include, re_path

from .sites import dashboard, website
from .utils import decorator_include

urlpatterns = [
    re_path(r"^", include(website.urls)),
    re_path(
        r"^dashboard/",
        decorator_include([login_required], dashboard.urls),
    ),
]



import logging

from django.contrib import messages
from django.contrib.admin.utils import quote
from django.shortcuts import redirect

# from django.contrib.auth.decorators import login_required
# from django.core.exceptions import PermissionDenied
# from django.http.response import HttpResponseRedirect
# from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views.generic.base import ContextMixin, TemplateResponseMixin, TemplateView, View

# from django.views.generic.detail import BaseDetailView, SingleObjectMixin
# from django.views.generic.edit import BaseCreateView, BaseDeleteView, BaseUpdateView
from django.views.generic.detail import SingleObjectMixin, SingleObjectTemplateResponseMixin
from django.views.generic.list import MultipleObjectMixin

import simpel_hookup.core as hookup  # NOQA
from django_filters.views import FilterMixin

before_inspect_hook_name = "BEFORE_INSPECT_VIEW_HOOK"
after_inspect_hook_name = "AFTER_INSPECT_VIEW_HOOK"

logger = logging.getLogger("simpel")


class SiteContext(ContextMixin):
    title = ""
    subtitle = ""
    meta_title = None
    meta_description = None

    def get_title(self):
        return self.title

    def get_subtitle(self):
        return self.subtitle

    def get_meta_title(self):
        return self.meta_title or self.get_title()

    def get_meta_description(self):
        return self.meta_description or self.get_title()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "view": self,
                "title": self.get_title(),
                "subtitle": self.get_subtitle(),
                "meta_title": self.get_meta_title(),
                "meta_decription": self.get_meta_description(),
            }
        )
        return context


class BaseView(TemplateResponseMixin, SiteContext, View):
    def __init__(self, title="", subtitle="", meta_title=None, meta_description=None, extra_context=None, **kwargs):
        self.title = title
        self.subtitle = subtitle
        self.meta_title = meta_title
        self.meta_description = meta_description
        super().__init__(extra_context=extra_context, **kwargs)

    def has_permission(self, request):
        pass

    def dispatch(self, request, *args, **kwargs):
        self.has_permission(request)
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)


class SiteView(BaseView):
    viewset = None
    model = None

    def __init__(self, viewset, **kwargs):
        self.viewset = viewset
        self.namespace = viewset.namespace
        self.url_helper = viewset.url_helper
        self.permission_helper = viewset.permission_helper
        self.model = getattr(self.viewset, "model", None)
        self.opts = self.model._meta
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "model": self.model,
                "opts": self.opts,
            }
        )
        return context


class ModelView(SiteView):
    model = None

    def __init__(self, viewset, **kwargs):
        self.model = viewset.model
        super().__init__(viewset, **kwargs)


class ListView(FilterMixin, MultipleObjectMixin, ModelView):
    def __init__(self, viewset, **kwargs):
        self.queryset = viewset.get_queryset()
        self.ordering = viewset.get_ordering()
        self.paginator_class = viewset.get_paginator_class()
        self.paginate_by = viewset.paginate_by
        self.paginate_orphans = viewset.paginate_orphans
        self.filterset_class = viewset.get_filterset_class()
        self.filterset_fields = viewset.get_filterset_fields()
        super().__init__(viewset, **kwargs)

    @cached_property
    def index_url(self):
        return self.url_helper.get_url("index")

    def get_template_names(self):
        return self.viewset.get_templates(action="index")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        return context

    def get(self, request, *args, **kwargs):
        filterset_class = self.get_filterset_class()
        self.filterset = self.get_filterset(filterset_class)

        if not self.filterset.is_bound or self.filterset.is_valid() or not self.get_strict():
            self.object_list = self.filterset.qs
        else:
            self.object_list = self.filterset.queryset.none()

        context = self.get_context_data(filter=self.filterset, object_list=self.object_list)
        return self.render_to_response(context)


class InstanceSpecificMixin(SingleObjectMixin):
    """A base view for displaying a single object."""

    def __init__(self, viewset, **kwargs):
        self.queryset = viewset.get_queryset()
        super().__init__(viewset, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().dispatch(request, *args, **kwargs)

    def get_page_subtitle(self):
        return self.object

    @cached_property
    def index_url(self):
        return self.url_helper.get_url("index")

    @cached_property
    def edit_url(self):
        return self.url_helper.get_url("edit", quote(self.object.id))

    @cached_property
    def delete_url(self):
        return self.url_helper.get_url("delete", quote(self.object.id))

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(object=self.object)
        return self.render_to_response(context)


class InspectView(SingleObjectTemplateResponseMixin, InstanceSpecificMixin, ModelView):
    """A view for displaying a object detail."""

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({"action": "inspect"})
        return context

    def check_action_permitted(self, user):
        return self.permission_helper.user_can_inspect_obj(user, self.object) or self.viewset.inspect_public

    def get_template_names(self):
        """
        Return a list of inspect template names defined in viewset.
        """
        return self.viewset.get_inspect_template()


class DeleteView(InspectView):
    """A view for displaying an object deletion view."""

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({"action": "delete"})
        return context

    def check_action_permitted(self, user):
        return self.permission_helper.user_can_delete_obj(user, self.object)

    def get_template_names(self):
        """
        Return a list of delete template names defined in viewset.
        """
        return self.viewset.get_delete_template()

    def get_success_message(self):
        return _("%s deleted!") % self.object

    def post(self, request, *args, **kwargs):
        try:
            self.object.delete()
            msg = self.get_success_message()
            messages.success(request, msg)
        except Exception as err:
            logger.error(err)
            messages.error(request, _("Object deletion failed!"))
        return redirect(self.index_url)


class Error404View(TemplateView):
    template_name = "404.html"


class Error403View(TemplateView):
    template_name = "403.html"


class Error500View(TemplateView):
    template_name = "500.html"


class MaintenanceView(TemplateView):
    template_name = "maintenance.html"



from importlib import import_module

from django.urls import URLPattern, URLResolver, include
from django.utils.functional import cached_property


class DecoratedPatterns(object):
    """
    A wrapper for an urlconf that applies a decorator to all its views.
    """

    def __init__(self, urlconf_module, decorators):
        # ``urlconf_module`` may be:
        #   - an object with an ``urlpatterns`` attribute
        #   - an ``urlpatterns`` itself
        #   - the dotted Python path to a module with an ``urlpatters`` attribute
        self.urlconf = urlconf_module
        try:
            iter(decorators)
        except TypeError:
            decorators = [decorators]
        self.decorators = decorators

    def decorate_pattern(self, pattern):
        if isinstance(pattern, URLResolver):
            decorated = URLResolver(
                pattern.pattern,
                DecoratedPatterns(pattern.urlconf_module, self.decorators),
                pattern.default_kwargs,
                pattern.app_name,
                pattern.namespace,
            )
        else:
            callback = pattern.callback
            for decorator in reversed(self.decorators):
                callback = decorator(callback)
            decorated = URLPattern(
                pattern.pattern,
                callback,
                pattern.default_args,
                pattern.name,
            )
        return decorated

    @cached_property
    def urlpatterns(self):
        # urlconf_module might be a valid set of patterns, so we default to it.
        patterns = getattr(self.urlconf_module, "urlpatterns", self.urlconf_module)
        return [self.decorate_pattern(pattern) for pattern in patterns]

    @cached_property
    def urlconf_module(self):
        if isinstance(self.urlconf, str):
            return import_module(self.urlconf)
        else:
            return self.urlconf

    @cached_property
    def app_name(self):
        return getattr(self.urlconf_module, "app_name", None)


def decorator_include(decorators, arg, namespace=None):
    """
    A replacement for ``django.conf.urls.include`` that takes a decorator,
    or an iterable of view decorators as the first argument and applies them, in
    reverse order, to all views in the included urlconf.

    from django.contrib import admin
    from django.core.exceptions import PermissionDenied
    from django.urls import path
    from django.contrib.auth.decorators import login_required, user_passes_test

    from decorator_include import decorator_include

    from mysite.views import index

    def only_user(username):
        def check(user):
            if user.is_authenticated and user.username == username:
                return True
            raise PermissionDenied
        return user_passes_test(check)

    urlpatterns = [
        path('', views.index, name='index'),
        # will redirect to login page if not authenticated
        path('secret/', decorator_include(login_required, 'mysite.secret.urls')),
        # will redirect to login page if not authenticated
        # will return a 403 http error if the user does not have the "god" username
        path('admin/', decorator_include([login_required, only_user('god')], admin.site.urls),
    ]

    """
    if isinstance(arg, tuple) and len(arg) == 3 and not isinstance(arg[0], str):
        # Special case where the function is used for something like `admin.site.urls`, which
        # returns a tuple with the object containing the urls, the app name, and the namespace
        # `include` does not support this pattern (you pass directly `admin.site.urls`, without
        # using `include`) but we have to
        urlconf_module, app_name, namespace = arg
    else:
        urlconf_module, app_name, namespace = include(arg, namespace=namespace)
    return DecoratedPatterns(urlconf_module, decorators), app_name, namespace



from functools import wraps


def maintenance(test_func):
    """
    Decorator for views that checks whether a site is in maintenance mode.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            from simpel_routers.views import MaintenanceView

            if not test_func(request.user):
                return view_func(request, *args, **kwargs)
            else:
                return MaintenanceView.as_view()(request, *args, **kwargs)

        return _wrapped_view

    return decorator


def maintainer_required(function=None):
    """
    Decorator for views that checks that the user is logged in, redirecting
    to the log-in page if necessary.
    """
    actual_decorator = maintenance(lambda u: not u.is_superuser)
    if function:
        return actual_decorator(function)
    return actual_decorator



"""
Routers provide a convenient and consistent way of automatically
determining the URL conf for your Web.

They are used by simply instantiating a Router class, and then registering
all the required ViewSets with that router.

For example, you might have a `urls.py` that looks something like this:

    router = DefaultRouter()
    router.register(UserViewSet)
    router.register(AccountViewsetGroup)
    urlpatterns = router.urls

"""

from audioop import reverse
from inspect import isclass
from logging import getLogger

from django.core.exceptions import ImproperlyConfigured
from django.urls.conf import include, path
from django.views import View

import simpel_hookup.core as hookup

from .views import BaseView

logger = getLogger("simpel")


class DefaultIndexView(BaseView):
    template_name = "index.html"


class BaseRouter:

    namespace = None

    def __init__(self):
        self.registry = []
        if self.namespace is None:
            raise ImproperlyConfigured(
                "%s router namespace required!" % self.__class__.__name__,
            )

    def register(self, viewset_class):
        viewset = viewset_class(router=self)
        self.registry.append(viewset)

        # invalidate the urls cache
        if hasattr(self, "_urls"):
            del self._urls

    def get_urls(self):
        """
        Return a list of URL patterns, given the registered viewsets.
        """
        raise NotImplementedError("get_urls must be overridden")

    @property
    def urls(self):
        if not hasattr(self, "_urls"):
            self._urls = self.get_urls()
        return self._urls


class SimpleRouter(BaseRouter):
    def get_urls(self):
        """
        Use the registered viewsets to generate a list of URL patterns.
        """
        urls = []
        for viewset in self.registry:
            urls.append(path("%s" % viewset.get_prefix(), include(viewset.urls)))
        return urls


class DefaultRouter(SimpleRouter):
    """
    The default router extends the SimpleRouter, but also adds in a default
    API root view, and adds format suffix patterns to the URLs.
    """

    index_enabled = True
    index_view_name = "index"
    index_template_name = "index.html"
    index_view_class = DefaultIndexView
    site_title = "Simpel Website"
    site_header = "Simpel Website"
    site_view_hook_name = "REGISTER_WEBSITE_VIEW"
    site_path_hook_name = "REGISTER_WEBSITE_PATH"

    def each_context(self, request):
        """
        Return a dictionary of variables to put in the template context for
        *every* page in this router site.
        """
        return {
            "site_title": self.site_title,
            "site_header": self.site_header,
            "site_url": reverse("%s_%s" % (self.namespace, self.index_view_name)),
        }

    def get_index_view_class(self):
        """
        Return a basic root view.
        """
        return self.index_view_class

    def get_hooked_views(self):
        # Get registered custom admin view
        funcs = hookup.get_hooks(self.site_view_hook_name)
        urls = []
        for func in funcs:
            url_path, view, name = func()
            if isclass(view):
                if not issubclass(view, View):
                    raise ImproperlyConfigured("%s must be subclass of View" % view)
                route = path(url_path, view.as_view(), name=name)
            elif callable(view):
                route = path(url_path, view, name=name)
            else:
                raise ImproperlyConfigured("%s must be View or function" % view)
            urls.append(route)
        return urls

    def get_hooked_paths(self):
        urls = []
        funcs = hookup.get_hooks(self.site_path_hook_name)
        for func in funcs:
            urls.append(func())
        return urls

    def get_urls(self):
        """
        Generate the list of URL patterns, including a default root view
        for the API, and appending `.json` style format suffixes.
        """
        urls = super().get_urls()
        if self.index_enabled:
            urls += (
                path(
                    "",
                    self.get_index_view_class().as_view(template_name=self.index_template_name),
                    name=self.index_view_name,
                ),
            )
        urls += self.get_hooked_views()
        urls += self.get_hooked_paths()
        return urls



from django.contrib.auth import views
from django.urls import path
from django.utils.translation import gettext_lazy as _

import simpel_hookup.core as hookup

from .routers import DefaultIndexView, DefaultRouter, SimpleRouter
from .settings import routers_settings


class IndexView(DefaultIndexView):
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_template_names(self):
        return super().get_template_names()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({"title": _("Simpellab")})
        return context


class AuthenticationRouter(SimpleRouter):
    namespace = "account"

    def get_urls(self):

        urls = super().get_urls()
        urls += [
            path("login/", views.LoginView.as_view(redirect_authenticated_user=True), name="login"),
            path("logout/", views.LogoutView.as_view(), name="logout"),
            path("password_reset/", views.PasswordResetView.as_view(), name="password_reset"),
            path("password_reset/done/", views.PasswordResetDoneView.as_view(), name="password_reset_done"),
            path("reset/<uidb64>/<token>/", views.PasswordResetConfirmView.as_view(), name="password_reset_confirm"),
            path("reset/done/", views.PasswordResetCompleteView.as_view(), name="password_reset_complete"),
            path("password_change/", views.PasswordChangeView.as_view(), name="password_change"),
            path("password_change/done/", views.PasswordChangeDoneView.as_view(), name="password_change_done"),
        ]
        return urls


class DashboardRouter(DefaultRouter):
    namespace = "dashboard"
    index_view_class = IndexView
    site_title = routers_settings.DASHBOARD_TITLE
    site_header = routers_settings.DASHBOARD_HEADER
    index_template_name = "dashboard/index.html"
    site_view_hook_name = "REGISTER_DASHBOARD_VIEW"
    site_path_hook_name = "REGISTER_DASHBOARD_PATH"


class WebsiteRouter(DefaultRouter):
    namespace = "website"
    index_template_name = "website/index.html"
    index_enabled = routers_settings.WEBSITE_INDEX_ENABLED
    site_title = routers_settings.WEBSITE_TITLE
    site_header = routers_settings.WEBSITE_HEADER
    site_view_hook_name = "REGISTER_WEBSITE_VIEW"
    site_path_hook_name = "REGISTER_WEBSITE_PATH"


def get_router(router_class, hook_name):
    router = router_class()
    viewsets = [func() for func in hookup.get_hooks(hook_name)]
    for viewset in viewsets:
        router.register(viewset)
    return router


website = get_router(WebsiteRouter, "REGISTER_WEBSITE_VIEWSET")
dashboard = get_router(DashboardRouter, "REGISTER_DASHBOARD_VIEWSET")
